/* 
** File: c_cvaroverrides.cpp
**
** Written By: Chris Laverdure <dashevil@sympatico.ca>
**
** This file belongs to the public domain
*/

#include <stdio.h>
#include <string.h>
#include "c_console.h"
#include "c_dispatch.h"
#include "w_wad.h"
#include "c_cvars.h"
#include "doomtype.h"
#include "c_cvaroverrides.h"

static cvaroverride_t *override_list = NULL;
static char overrides_effective = 0;
static char overrides_map[SIZE_NAME];

static cvaroverride_t *LL_search (cvaroverride_t *LL_startpoint, char *LL_mapname)
{

   cvaroverride_t *__lastlink;

   __lastlink = LL_startpoint;

   while (__lastlink->next != NULL && !(stricmp(__lastlink->name, LL_mapname) == 0))
	__lastlink = __lastlink->next;

   return __lastlink;
}

CCMD (add_cvaroverride)
{

   cvaroverride_t *pivot = NULL;

   if (overrides_effective)
   {
	Printf("Error: You cannot specify CVar Overrides while some are active.\n");
	return;
   }

   if (argv.argc() != 4)
   { // Obviously they don't know how to use it.
      Printf("Error: Invalid number of parameters!\nUsage:\nadd_cvaroverride (map name) (cvar name) (value)\n");
      return;
   }

   // Does this map exist?
   if (W_CheckNumForName (argv[1]) == -1)
   { // No it does not.
      Printf("Error: Map `%s` doesn't exist.\n", argv[1]);
      return;
   }

   // Is it a valid cvar?
   FBaseCVar *var = FindCVarSub (argv[2], strlen(argv[2]));
   if (var == NULL)
   { // It's not.
      Printf("Error: Cvar `%s` doesn't exist.\n", argv[2]);
      return;
   }
   else if (var->GetFlags() & CVAR_LATCH)
   { // It's latched, this isn't supported yet.
      Printf("Error: Cvar `%s` is latched, cannot override.\n", argv[2]);
      return;
   }

   UCVarValue val = var->GetGenericRep (CVAR_String);

   Printf("Override for %s: %s = `%s` -> `%s`\n", argv[1], argv[2],
						val.String, argv[3]);

   // Is there a list to add to, or are we making it?
   if (override_list == NULL)
   { // First entry
   
      override_list = (cvaroverride_t *) malloc(sizeof(cvaroverride_t));
      override_list->next = NULL;
      pivot = override_list;
   
   }
   else
   { // Add to the list

      // *sigh* We're going to need more memory for this.
      pivot = (cvaroverride_t *) malloc(sizeof(cvaroverride_t));

      // Since we plan on applying all the overrides for a map in one swoop
      // it is wise to place them all together so that we can make that
      // assumption later.

      // Step 1: Does the first override belong to the same map we do?
      //         If so, just snap in before it.
      // Step 2: Check the link list for overrides that belong to the same
      //         map we do. Once found, assume our place in line.
      // Step 3: There are no overrides that belong to the same map we do.
      //         We go to the end of the line.

      if (!stricmp(argv[1], override_list->name))
      { // The first entry in the list matches, just place our new one before it.

         pivot->next = override_list;
         override_list = pivot;
   
      }
      else
      { // Guess we get to look through the whole list to find pals.
         cvaroverride_t *temp_pointer;

	 temp_pointer = LL_search(override_list, argv[1]);

	 pivot->next = temp_pointer->next;
	 temp_pointer->next = pivot;	 

      }
   }

   // Set the values

   strncpy(pivot->name, argv[1], SIZE_NAME);
   strncpy(pivot->cvar, argv[2], SIZE_CVAR);
   strncpy(pivot->override, argv[3], SIZE_OVERRIDE);
   strncpy(pivot->oldvalue, val.String, SIZE_OLDVALUE);

}

CCMD (list_cvaroverride)
{
   cvaroverride_t *index;

   index = override_list;

   // Run through the link list and display the overrides.

   Printf("+-----+--------------------+--------------------------+--------------------------+\n");
   Printf("|Map  |Cvar                |Old Value                 |Override                  |\n");
   Printf("+-----+--------------------+--------------------------+--------------------------+\n");

   if (index == NULL)
   {
      Printf("[CVAR OVERRIDE LIST EMPTY]\n");
   }   

   while (index != NULL)
   {
      printf("|%5s|%20s|%26s|%26s|\n", index->name, index->cvar,
					index->oldvalue, index->override);
      index = index->next;
   }

   Printf("+-----+--------------------+--------------------------+--------------------------+\n");

}

void cvaroverride_revert(void)
{

   cvaroverride_t *index = NULL;

   if (!overrides_effective)
   { // DUCKS CAN'T SWIM IF THEY ARE SINKING?
	return;
   }

   if (override_list == NULL) // There are no cvar overrides.
      return;

   index = LL_search(override_list, overrides_map);

   Printf("Reverting previous CVAR overrides. [%s]\n", overrides_map);  

   if (index == NULL)
      printf("cvaroverride_revert: Index is null!\n"); 

   overrides_effective = 0;

   while (index != NULL && (stricmp(index->name, overrides_map) == 0))
   {
      FBaseCVar *var = FindCVarSub (index->cvar, strlen(index->cvar));

      var->CmdSet (index->oldvalue);

      Printf("CVAR REVERT: %s = `%s` -> `%s`\n", index->cvar, index->override, index->oldvalue);

      index = index->next;
   }
}

void cvaroverride_enable(char *mapname)
{

   cvaroverride_t *index = NULL;

   if (overrides_effective)
   { // WHAT THE FUCK?
	Printf("WARNING: void cvaroverride_enable called while previous overrides are still active\n");
	Printf("         Ignoring new overrides.\n");
	return;
   }

   if (override_list == NULL) // There are no cvar overrides.
	return;
   
   index = LL_search(override_list, mapname);

   if (!(stricmp(index->name, mapname) == 0)) // There are no cvar overrides for this map.
      return;

   overrides_effective = 1;
   strncpy(overrides_map, mapname, SIZE_NAME);

   Printf("Enabling CVAR overrides for this map. [%s]\n", mapname);

   while (index != NULL && (stricmp(index->name, mapname) == 0))
   {
      FBaseCVar *var = FindCVarSub (index->cvar, strlen(index->cvar));

      var->CmdSet (index->override);

      Printf("CVAR OVERRIDE: %s = `%s` -> `%s`\n", index->cvar, index->oldvalue, index->override);

      index = index->next; 

   }

}
