// Emacs style mode select	 -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//		DOOM selection menu, options, episode etc.
//		Sliders and icons. Kinda widget stuff.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <zlib.h>

#if _WIN32
#include <io.h>
#else
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include "doomdef.h"
#include "gstrings.h"
#include "c_console.h"
#include "c_dispatch.h"
#include "d_main.h"
#include "i_system.h"
#include "i_video.h"
#include "z_zone.h"
#include "v_video.h"
#include "w_wad.h"
#include "r_local.h"
#include "hu_stuff.h"
#include "g_game.h"
#include "m_argv.h"
#include "m_swap.h"
#include "m_random.h"
#include "s_sound.h"
#include "doomstat.h"
#include "m_menu.h"
#include "v_text.h"
#include "st_stuff.h"
#include "d_gui.h"
#include "version.h"

#include "lists.h"
#include "gi.h"

// TYPES -------------------------------------------------------------------

struct FSaveGameNode : public Node
{
	char Title[SAVESTRINGSIZE];
	char *Filename;
	bool bOldVersion;
};


// PRIVATE DATA DEFINITIONS ------------------------------------------------
static List				SaveGames;
static FSaveGameNode	*SelSaveGame;
static FSaveGameNode	NewSaveNode;

// CODE --------------------------------------------------------------------

static void M_InsertSaveNode (FSaveGameNode *node)
{
	FSaveGameNode *probe;

	if (SaveGames.IsEmpty ())
	{
		SaveGames.AddHead (node);
		return;
	}

	if (node->bOldVersion)
	{ // Add node at bottom of list
		probe = static_cast<FSaveGameNode *>(SaveGames.TailPred);
		while (probe->Pred != NULL && probe->bOldVersion &&
			stricmp (node->Title, probe->Title) < 0)
		{
			probe = static_cast<FSaveGameNode *>(probe->Pred);
		}
		node->Insert (probe);
	}
	else
	{ // Add node at top of list
		probe = static_cast<FSaveGameNode *>(SaveGames.Head);
		while (probe->Succ != NULL && !probe->bOldVersion &&
			stricmp (node->Title, probe->Title) > 0)
		{
			probe = static_cast<FSaveGameNode *>(probe->Succ);
		}
		node->InsertBefore (probe);
	}
}


//
// M_ReadSaveStrings
//
// Find savegames and read their titles
//
static void M_ReadSaveStrings ()
{
	if (SaveGames.IsEmpty ())
	{
		long filefirst;
		findstate_t c_file;
		char filter[PATH_MAX];

		G_BuildSaveName (filter, "*.zds", -1);
		filefirst = I_FindFirst (filter, &c_file);
		if (filefirst != -1)
		{
			do
			{
				// I_FindName only returns the file's name and not its full path
				char filepath[PATH_MAX];
				G_BuildSaveName (filepath, I_FindName(&c_file), -1);
				FILE *file = fopen (filepath, "rb");
				if (file != NULL)
				{
					char sig[PATH_MAX];

					if (fread (sig, 1, 16, file) == 16)
					{
						char title[SAVESTRINGSIZE];
						int i, c;
						bool oldVer = true;
						bool addIt = false;

						if (strncmp (sig, SAVESIG, 9) == 0)
						{
							fread (title, 1, SAVESTRINGSIZE, file);
							fseek (file, 8, SEEK_CUR);		// skip map name

							if (strncmp (sig, SAVESIG, 16) == 0)
							{
								// Was saved with a compatible ZDaemon version,
								// so check if it's for the current game.
								// If it is, add it. Otherwise, ignore it.
								i = 0;
								do
								{
									c = fgetc (file);
									sig[i++] = c;
								} while (c != EOF && c != 0);
								if (c != EOF && stricmp (sig, W_GetWadName (1)) == 0)
								{
									addIt = true;
									oldVer = false;
								}
							}
							else
							{ // Game is from a different ZDaemon version. Add it
							  // just so that the user has a means of deleting it
							  // if desired.
								addIt = true;
							}
						}
						else
						{ // Before 1.23 beta 21, the savesig came after the title
							memcpy (title, sig, 16);
							if (fread (title + 16, 1, SAVESTRINGSIZE-16, file) == SAVESTRINGSIZE-16 &&
								fread (sig, 1, 16, file) == 16 &&
								strncmp (sig, SAVESIG, 9) == 0)
							{ // Game is from an old ZDoom.
								addIt = true;
							}
						}

						if (addIt)
						{
							FSaveGameNode *node = new FSaveGameNode;
							node->Filename = copystring (filepath);
							node->bOldVersion = oldVer;
							memcpy (node->Title, title, SAVESTRINGSIZE);
							M_InsertSaveNode (node);
						}
					}
					fclose (file);
				}
			} while (I_FindNext (filefirst, &c_file) == 0);
			I_FindClose (filefirst);
		}
	}
	if (SelSaveGame == NULL || SelSaveGame->Succ == NULL)
	{
		SelSaveGame = static_cast<FSaveGameNode *>(SaveGames.Head);
	}
}

void M_NotifyNewSave (const char *file, const char *title, bool okForQuicksave)
{
	FSaveGameNode *node;

	M_ReadSaveStrings ();

	// See if the file is already in our list
	for (node = static_cast<FSaveGameNode *>(SaveGames.Head);
		 node->Succ != NULL;
		 node = static_cast<FSaveGameNode *>(node->Succ))
	{
#ifdef unix
		if (strcmp (node->Filename, file) == 0)
#else
		if (stricmp (node->Filename, file) == 0)
#endif
		{
			strcpy (node->Title, title);
			node->bOldVersion = false;
			break;
		}
	}

	if (node->Succ == NULL)
	{
		node = new FSaveGameNode;
		strcpy (node->Title, title);
		node->Filename = copystring (file);
		node->bOldVersion = false;
		M_InsertSaveNode (node);
		SelSaveGame = node;
	}

}

//
// M_Init
//
void M_Init (void)
{
	strcpy (NewSaveNode.Title, "<New Save Game>");
}

